from __future__ import annotations

import copy
from dataclasses import _MISSING_TYPE, dataclass, field, fields
from datetime import datetime
from enum import Enum
from pathlib import Path
from typing import Dict, List, Optional

from slugify import slugify

from .exceptions import InvalidConfiguration


class LogLevel(Enum):
    info = "INFO"
    debug = "DEBUG"
    warning = "WARNING"


class MountPointType(Enum):
    HOSTPATH = "HOSTPATH"
    SECRET = "SECRET"
    CONFIGMAP = "CONFIGMAP"
    EMPTYDIR = "EMPTYDIR"


@dataclass
class MountPoint:
    local_path: str
    mount_path: str
    read_only: bool = True
    type: MountPointType = MountPointType.HOSTPATH
    size_limit: Optional[str] = None  # Only valid for emptydir

    @property
    def name(self) -> str:
        return slugify(self.local_path)


def create_klass(cls, settings, prefix: str):
    """
    Create cls from config parameters in settings.


    The fields of cls should be present in upper-case prefixed with the prefix and "_".
    """
    kwargs = {}

    for cls_field in fields(cls):
        attr_name = f"{prefix}_{cls_field.name.upper()}"

        if (
            isinstance(cls_field.default, _MISSING_TYPE)
            and isinstance(cls_field.default_factory, _MISSING_TYPE)
            and not hasattr(settings, attr_name)
        ):
            raise InvalidConfiguration(f"{attr_name} is a mandatory setting")

        if hasattr(settings, attr_name):
            kwargs[cls_field.name] = getattr(settings, attr_name)

    # [['api', 'CLUSTER', 'v3'], ]
    if "constraints" in kwargs and isinstance(kwargs["constraints"], list):
        kwargs["constraints"] = [MarathonConstraint(*x) for x in kwargs["constraints"]]

    return cls(**kwargs)


@dataclass
class DockerBackendConfig:
    client_url: str  # URL to docker or marathon
    network_name: str  # network to use
    debug: bool = False

    @classmethod
    def from_settings(cls, settings, prefix="CONTAINER_HUB") -> "DockerBackendConfig":
        """
        Populate DockerBackendConfig from simple-settings, Django settings
        or similair object.
        """
        return create_klass(cls, settings, prefix)


@dataclass
class MarathonConstraint:
    param: str
    operator: str
    value: str


@dataclass
class MarathonBackendConfig(DockerBackendConfig):
    constraints: List[MarathonConstraint] = field(default_factory=list)

    @classmethod
    def from_settings(cls, settings, prefix="CONTAINER_HUB") -> "MarathonBackendConfig":
        """
        Populate MarathonBackendConfig from simple-settings, Django settings
        or similair object.
        """
        return create_klass(cls, settings, prefix)


@dataclass
class KubernetesBackendConfig:
    client_url: str  # URL to docker or marathon
    # Kubernetes namespace to deploy to
    namespace: str = "default"

    @classmethod
    def from_settings(
        cls, settings, prefix="CONTAINER_HUB"
    ) -> "KubernetesBackendConfig":
        """
        Populate KubernetesBackendConfig from simple-settings, Django settings
        or similair object.
        """
        return create_klass(cls, settings, prefix)


@dataclass
class EnvVar:
    name: str
    value: str


@dataclass
class Label:
    name: str
    value: str


@dataclass
class ContainerConfig:
    image_name: str  # threedicore image name
    base_result_path: Path  # path for results
    sim_uid: str
    sim_ref_datetime: datetime
    end_time: int
    duration: int
    pause_timeout: int
    start_mode: str
    model_config: str
    max_cpu: int  # max CPU's to use
    session_memory: int  # max memory
    envs: List[EnvVar]
    labels: List[Label]
    max_rate: float = 0.0
    clean_up_files: bool = False
    gridadmin_download_url: Optional[str] = None
    tables_download_url: Optional[str] = None
    mount_points: List[MountPoint] = field(default_factory=list)
    redis_host: str = "redis"  # Local redis host
    container_log_level: LogLevel = LogLevel.info
    debugpy_port: Optional[int] = None


# Kubernetes dataclasses.
# These are much more generic than the Docker/Marathon ContainerConfig


@dataclass
class ResourceSpec:
    cpu: str
    memory: str

    @classmethod
    def from_dict(cls, values: Dict) -> "ResourceSpec":
        # Make (deep) copy to keep original defaults
        values = copy.deepcopy(values)

        return ResourceSpec(cpu=values["cpu"], memory=values["memory"])

    def to_dict(self):
        return {"cpu": self.cpu, "memory": self.memory}


@dataclass
class NodeAffinityConfig:
    key: str
    operator: str
    values: List[str]

    @classmethod
    def from_dict(cls, values: Dict) -> "NodeAffinityConfig":
        # Make (deep) copy to keep original defaults
        values = copy.deepcopy(values)

        return NodeAffinityConfig(
            key=values["key"], operator=values["operator"], values=values["values"]
        )


@dataclass
class Resources:
    limits: ResourceSpec
    requests: ResourceSpec

    @classmethod
    def from_dict(cls, values: Dict) -> "Resources":
        # Make (deep) copy to keep original defaults
        values = copy.deepcopy(values)

        return Resources(
            limits=ResourceSpec.from_dict(values.get("limits", {})),
            requests=ResourceSpec.from_dict(values.get("requests", {})),
        )


@dataclass
class KubernetesContainer:
    name: str
    image: str  # Docker image
    args: List[str]  # command start up args
    envs: List[EnvVar] = field(default_factory=list)
    labels: List[Label] = field(default_factory=list)
    mount_points: List[MountPoint] = field(default_factory=list)
    ports: List[int] = field(default_factory=list)
    resources: Optional[Resources] = None

    @classmethod
    def from_dict(cls, values: Dict) -> "KubernetesContainer":
        # Make (deep) copy to keep original defaults
        values = copy.deepcopy(values)

        return KubernetesContainer(
            name=values.get("name"),
            image=values.get("image"),
            args=values.get("args"),
            envs=[EnvVar(key, value) for key, value in values.get("envs", {}).items()],
            labels=[
                Label(key, value) for key, value in values.get("labels", {}).items()
            ],
            mount_points=[
                MountPoint(
                    key,
                    value["bind"],
                    value["ro"],
                    MountPointType(value.get("type", MountPointType.HOSTPATH.value)),
                    size_limit=value.get("size_limit", None),
                )
                for key, value in values.get("mounts", {}).items()
            ],
            ports=values.get("ports", []),
            resources=Resources.from_dict(values["resources"])
            if "resources" in values
            else None,
        )


@dataclass
class HostAlias:
    ip_address: str
    hostnames: List[str]


@dataclass
class KubernetesJobConfig:
    name: str
    redis_config: KubernetesContainer
    scheduler_config: KubernetesContainer
    simulation_config: KubernetesContainer
    annotations: List[Label] = field(default_factory=list)
    labels: List[Label] = field(default_factory=list)
    host_aliases: List[HostAlias] = field(default_factory=list)
    service_account_name: str = "simulation-service-account"
    regcred_secret_name: Optional[str] = None  # Registry credentials secret
    node_affinity: Optional[NodeAffinityConfig] = None

    @property
    def mount_points(self) -> List[MountPoint]:
        """
        Combined list of mount_points
        """
        mount_points = []
        for mount_point in (
            self.redis_config.mount_points
            + self.scheduler_config.mount_points
            + self.simulation_config.mount_points
        ):
            if mount_point not in mount_points:
                mount_points.append(mount_point)
        return mount_points

    @classmethod
    def from_settings(
        cls, name: str, settings, prefix="CONTAINER_HUB"
    ) -> "KubernetesJobConfig":
        """
        Load KubernetesJobConfig from supplied values in `CONTAINER_HUB_KUBERNETES_CONTAINER_DEFAULTS`
        """
        cfg = getattr(settings, f"{prefix}_KUBERNETES_CONTAINER_DEFAULTS")

        return KubernetesJobConfig(
            name=name,
            redis_config=KubernetesContainer.from_dict(cfg["REDIS"]),
            scheduler_config=KubernetesContainer.from_dict(cfg["SCHEDULER"]),
            simulation_config=KubernetesContainer.from_dict(cfg["SIMULATION"]),
            host_aliases=[
                HostAlias(ip_address, hostnames)
                for ip_address, hostnames in cfg.get("HOST_ALIASES", {}).items()
            ],
            annotations=[
                Label(name, value) for name, value in cfg.get("ANNOTATIONS", {}).items()
            ],
            regcred_secret_name=cfg.get("regcred_secret_name", None),
            node_affinity=NodeAffinityConfig(**cfg["node_affinity"])
            if "node_affinity" in cfg
            else None,
        )



from functools import cached_property
from typing import Any, Dict, List, Union

from hikaru.model.rel_1_20 import (
    Affinity,
    ConfigMapVolumeSource,
    Container,
    ContainerPort,
    EmptyDirVolumeSource,
    EnvVar,
    HostAlias,
    HostPathVolumeSource,
    Job,
    JobSpec,
    LocalObjectReference,
    NodeAffinity,
    NodeSelector,
    NodeSelectorRequirement,
    NodeSelectorTerm,
    ObjectMeta,
    PodSpec,
    PodTemplateSpec,
    ResourceRequirements,
    SecretVolumeSource,
    Volume,
    VolumeMount,
)
from kubernetes import config as client_config
from kubernetes.client import (
    ApiClient,
    ApiException,
    BatchV1Api,
    Configuration,
    CoreV1Api,
)
from kubernetes.client.models import V1DeleteOptions, V1Job, V1JobList, V1JobStatus

from container_hub.exceptions import CarrierError
from container_hub.models import (
    KubernetesBackendConfig,
    KubernetesContainer,
    KubernetesJobConfig,
    MountPointType,
    NodeAffinityConfig,
)

# Accessing host (laptop) from within k8s cluster
K3S_HOST_DNS_NAME = "host.k3d.internal"
MINIKUBE_HOST_DNS_NAME = "host.minikube.internal"


class KubernetesBackend:
    """
    Backend for starting Docker instances via Docker
    """

    def __init__(self, config: KubernetesBackendConfig, in_cluster=True):
        self.config = config
        self.in_cluster = in_cluster

    @cached_property
    def configuration(self) -> Configuration:
        configuration = Configuration(host=self.config.client_url)
        if self.in_cluster:
            client_config.load_incluster_config(configuration)

        return configuration

    def container_hosts(self) -> Dict[str, str]:
        return {}

    def get_job_status(self, job_name: str) -> Dict[str, Any]:
        """
        Get a specific pod ip, based on job_name (simulation-xxx)
        """
        job: Union[V1Job, None] = None
        with ApiClient(self.configuration) as api_client:
            api = BatchV1Api(api_client)
            try:
                job: V1Job = api.read_namespaced_job(
                    job_name, namespace=self.config.namespace
                )
            except ApiException:
                job = None

        if job is None:
            raise CarrierError("unknown Job")

        if job.status is None:
            raise CarrierError("Job has no status")

        job_status: V1JobStatus = job.status
        return job_status.to_dict()

    def pod_ip(self, job_name: str) -> str:
        """
        Get a specific pod ip, based on job_name (simulation-xxx)
        """
        with ApiClient(self.configuration) as api_client:
            api = CoreV1Api(api_client)
            results = api.list_namespaced_pod(
                namespace=self.config.namespace, label_selector=f"job-name={job_name}"
            )
            if len(results.items) != 1:
                raise CarrierError(f"Incorrect number of pods returned: {results}")
            return results.items[0].status.pod_ip

    def container_ips(self) -> Dict[str, str]:
        return {}

    def container_list(self) -> List[str]:
        with ApiClient(self.configuration) as api_client:
            jobs: List[Job] = []
            api = BatchV1Api(api_client)
            job_list = None
            while job_list is None or job_list.metadata._continue is not None:
                job_list: V1JobList = api.list_namespaced_job(
                    namespace=self.config.namespace
                )
                jobs += [x for x in job_list.items]
        return [
            job.metadata.name.lstrip("simulation-")
            for job in jobs
            if job.metadata.name.startswith("simulation-")
        ]

    def up(self, job_config: KubernetesJobConfig) -> str:
        """
        Create Kubernetes job for simulation
        """
        job = get_simulation_job(job_config)
        with ApiClient(self.configuration) as api_client:
            job = job.create(namespace=self.config.namespace, client=api_client)
        return job_config.name

    def down(self, sim_uid: str):
        """Remove the given app."""
        name = f"simulation-{sim_uid}"
        with ApiClient(self.configuration) as api_client:
            api = BatchV1Api(api_client)
            api.delete_namespaced_job(
                namespace=self.config.namespace,
                name=name,
                grace_period_seconds=0,
                propagation_policy="Background",
                body=V1DeleteOptions(propagation_policy="Background"),
            )


def get_node_affinity(node_affinity_cfg: NodeAffinityConfig) -> Affinity:
    return Affinity(
        nodeAffinity=NodeAffinity(
            requiredDuringSchedulingIgnoredDuringExecution=NodeSelector(
                nodeSelectorTerms=[
                    NodeSelectorTerm(
                        matchExpressions=[
                            NodeSelectorRequirement(
                                key=node_affinity_cfg.key,
                                operator=node_affinity_cfg.operator,
                                values=node_affinity_cfg.values,
                            )
                        ]
                    )
                ]
            )
        )
    )


def get_simulation_job(cfg: KubernetesJobConfig) -> Job:
    return Job(
        apiVersion="batch/v1",
        kind="Job",
        metadata=ObjectMeta(
            name=cfg.name,
            annotations={x.name: x.value for x in cfg.annotations},
            labels={"app": cfg.name},
        ),
        spec=JobSpec(
            template=PodTemplateSpec(
                metadata=ObjectMeta(
                    annotations={x.name: x.value for x in cfg.annotations},
                    labels={"app": cfg.name},
                ),
                spec=PodSpec(
                    affinity=get_node_affinity(cfg.node_affinity)
                    if cfg.node_affinity is not None
                    else None,
                    serviceAccountName=cfg.service_account_name,
                    imagePullSecrets=[
                        LocalObjectReference(name=cfg.regcred_secret_name)
                    ]
                    if cfg.regcred_secret_name is not None
                    else None,
                    hostAliases=[
                        HostAlias(x.ip_address, x.hostnames) for x in cfg.host_aliases
                    ],
                    containers=[
                        get_container(cfg.redis_config),
                        get_container(cfg.scheduler_config),
                        get_container(cfg.simulation_config),
                    ],
                    volumes=[
                        Volume(
                            name=mount.name,
                            hostPath=HostPathVolumeSource(
                                path=mount.local_path, type="Directory"
                            ),
                        )
                        for mount in cfg.mount_points
                        if mount.type == MountPointType.HOSTPATH
                    ]
                    + [
                        Volume(
                            name=mount.name,
                            emptyDir=EmptyDirVolumeSource(sizeLimit=mount.size_limit),
                        )
                        for mount in cfg.mount_points
                        if mount.type == MountPointType.EMPTYDIR
                    ]
                    + [
                        Volume(
                            name=mount.name,
                            configMap=ConfigMapVolumeSource(name=mount.local_path),
                        )
                        for mount in cfg.mount_points
                        if mount.type == MountPointType.CONFIGMAP
                    ]
                    + [
                        Volume(
                            name=mount.name,
                            secret=SecretVolumeSource(secretName=mount.local_path),
                        )
                        for mount in cfg.mount_points
                        if mount.type == MountPointType.SECRET
                    ],
                    restartPolicy="Never",
                ),
            ),
            backoffLimit=1,
            completions=1,
        ),
    )


def get_container(cfg: KubernetesContainer) -> Container:
    """
    Get k8s container API resource
    """
    # Might ne needed for local dev:
    #   hostAliases=[HostAlias(f"{HOST_IP}", ["minio"])],

    return Container(
        name=cfg.name,
        image=cfg.image,
        imagePullPolicy="IfNotPresent",
        resources=ResourceRequirements(
            limits=cfg.resources.limits.to_dict(),
            requests=cfg.resources.requests.to_dict(),
        )
        if cfg.resources is not None
        else ResourceRequirements(),
        args=cfg.args,
        env=[
            EnvVar(
                name=envvar.name,
                value=envvar.value,
            )
            for envvar in cfg.envs
        ],
        volumeMounts=[
            VolumeMount(
                mountPath=mount.mount_path, name=mount.name, readOnly=mount.read_only
            )
            for mount in cfg.mount_points
        ],
        ports=[ContainerPort(containerPort=port) for port in cfg.ports],
    )



from .exceptions import InvalidConfiguration


def get_backend(settings, prefix="CONTAINER_HUB"):
    """
    Get a backend based on simple-settings or Django settings object.

    Uses `CONTAINER_HUB_CARRIER` = ["docker", "kubernetes"] to determine
    the backend.
    """
    if not hasattr(settings, f"{prefix}_CARRIER"):
        raise InvalidConfiguration(f"{prefix}_CARRIER is a mandatory setting")
    opt = str(getattr(settings, f"{prefix}_CARRIER")).lower()

    if opt == "docker":
        from container_hub.carriers.docker.backend import (
            DockerBackend,
            DockerBackendConfig,
        )

        return DockerBackend(DockerBackendConfig.from_settings(settings, prefix))
    elif opt == "kubernetes":
        from container_hub.carriers.kubernetes.backend import (
            KubernetesBackend,
            KubernetesBackendConfig,
        )

        return KubernetesBackend(
            KubernetesBackendConfig.from_settings(settings, prefix)
        )

    raise InvalidConfiguration(
        f"Unknown carrier {opt} option, should either be 'docker' or 'marathon'"
    )



CONTAINER_HUB_CARRIER = "kubernetes"
CONTAINER_HUB_CLIENT_URL = "http://kubernetes_url/"
CONTAINER_HUB_NAMESPACE = "threedi"


# Note: Below are defaults that can always be overriden in code before
# sending it to Kubernetes
CONTAINER_HUB_KUBERNETES_CONTAINER_DEFAULTS = {
    "HOST_ALIASES": {
        "127.0.0.1": ["minio"],
    },
    "REDIS": {
        "name": "redis",
        "image": "redis:5.0.3-alpine",
        "args": ["sh", "-c", "rm -rf /data/dump.rdb && redis-server --save " ""],
        "ports": [
            6379,
        ],
    },
    "SCHEDULER": {
        "name": "scheduler",
        "image": "harbor.lizard.net/threedi/scheduler:latest",
        "args": ["python3", "/code/scheduler.py", "localhost"],
        "envs": {
            "DJANGO_SETTINGS_MODULE": "threedi_scheduler.developmentsettings",
            "REDIS_HOST": "localhost",
        },
        "mounts": {
            "/local/path/one": {"bind": "mount_path_1", "ro": True},
            "/local/path/two": {"bind": "mount_path_2", "ro": False},
            "scheduler-config": {
                "bind": "/etc/config/config.yaml",
                "ro": True,
                "type": "CONFIGMAP",
            },
            "scheduler-secret": {"bind": "/etc/secrets/", "ro": True, "type": "SECRET"},
            "shared-data": {
                "bind": "/mnt/results/",
                "ro": False,
                "type": "EMPTYDIR",
                "size_limit": "10Gi",
            },
        },
        "resources": {
            "limits": {
                "cpu": "1",
                "memory": "2Gi",
            },
            "requests": {"cpu": "250m", "memory": "128Mi"},
        },
    },
    "SIMULATION": {
        "name": "simulation",
        "image": "harbor.lizard.net/threedi/threedicore:2.16.1-2.2.5",
        "args": ["python", "service.py", "localhost"],
        "envs": {"RESULTS_PATH": "/results"},
        "mounts": {
            "/local/path/one": {"bind": "mount_path_1", "ro": True},
            "/local/path/two": {"bind": "mount_path_2", "ro": False},
            "shared-data": {
                "bind": "/mnt/results/",
                "ro": False,
                "type": "EMPTYDIR",
                "size_limit": "10Gi",
            },
        },
        "resources": {
            "limits": {
                "cpu": "1",
                "memory": "5Gi",
            },
            "requests": {"cpu": "1", "memory": "512Mi"},
        },
    },
    "regcred_secret_name": "regcred",
    "node_affinity": {
        "key": "is_compute",
        "operator": "in",
        "values": ["true"],
    },
}



from dataclasses import asdict

import pytest
from simple_settings import LazySettings

from container_hub import get_backend
from container_hub.carriers.docker.backend import DockerBackend
from container_hub.carriers.kubernetes.backend import KubernetesBackend
from container_hub.models import KubernetesJobConfig, MountPointType


@pytest.fixture
def docker_simple_settings():
    # os.environ.update({"SIMPLE_SETTINGS": "tests.test_files.docker_settings"})
    yield LazySettings("tests.test_files.docker_settings")


@pytest.fixture
def kubernetes_simple_settings():
    yield LazySettings("tests.test_files.kubernetes_settings")


def test_loading_docker_backend(docker_simple_settings):
    backend = get_backend(docker_simple_settings)
    assert isinstance(backend, DockerBackend)


def test_loading_kubernetes_backend(kubernetes_simple_settings):
    backend = get_backend(kubernetes_simple_settings)
    assert backend.config.namespace == "threedi"
    assert isinstance(backend, KubernetesBackend)


# Test loading kubernetes container defaults
def test_loading_kubernetes_defaults(kubernetes_simple_settings):
    config = KubernetesJobConfig.from_settings("sim4234", kubernetes_simple_settings)
    assert asdict(config) == {
        "host_aliases": [{"hostnames": ["minio"], "ip_address": "127.0.0.1"}],
        "name": "sim4234",
        "redis_config": {
            "name": "redis",
            "image": "redis:5.0.3-alpine",
            "args": ["sh", "-c", "rm -rf /data/dump.rdb && redis-server --save "],
            "envs": [],
            "labels": [],
            "mount_points": [],
            "ports": [6379],
            "resources": None,
        },
        "scheduler_config": {
            "name": "scheduler",
            "image": "harbor.lizard.net/threedi/scheduler:latest",
            "args": ["python3", "/code/scheduler.py", "localhost"],
            "envs": [
                {
                    "name": "DJANGO_SETTINGS_MODULE",
                    "value": "threedi_scheduler.developmentsettings",
                },
                {"name": "REDIS_HOST", "value": "localhost"},
            ],
            "labels": [],
            "mount_points": [
                {
                    "local_path": "/local/path/one",
                    "mount_path": "mount_path_1",
                    "read_only": True,
                    "size_limit": None,
                    "type": MountPointType.HOSTPATH,
                },
                {
                    "local_path": "/local/path/two",
                    "mount_path": "mount_path_2",
                    "read_only": False,
                    "size_limit": None,
                    "type": MountPointType.HOSTPATH,
                },
                {
                    "local_path": "scheduler-config",
                    "mount_path": "/etc/config/config.yaml",
                    "read_only": True,
                    "size_limit": None,
                    "type": MountPointType.CONFIGMAP,
                },
                {
                    "local_path": "scheduler-secret",
                    "mount_path": "/etc/secrets/",
                    "read_only": True,
                    "size_limit": None,
                    "type": MountPointType.SECRET,
                },
                {
                    "local_path": "shared-data",
                    "mount_path": "/mnt/results/",
                    "read_only": False,
                    "size_limit": "10Gi",
                    "type": MountPointType.EMPTYDIR,
                },
            ],
            "resources": {
                "limits": {
                    "cpu": "1",
                    "memory": "2Gi",
                },
                "requests": {"cpu": "250m", "memory": "128Mi"},
            },
            "ports": [],
        },
        "simulation_config": {
            "name": "simulation",
            "image": "harbor.lizard.net/threedi/threedicore:2.16.1-2.2.5",
            "args": ["python", "service.py", "localhost"],
            "envs": [{"name": "RESULTS_PATH", "value": "/results"}],
            "labels": [],
            "mount_points": [
                {
                    "local_path": "/local/path/one",
                    "mount_path": "mount_path_1",
                    "read_only": True,
                    "size_limit": None,
                    "type": MountPointType.HOSTPATH,
                },
                {
                    "local_path": "/local/path/two",
                    "mount_path": "mount_path_2",
                    "read_only": False,
                    "size_limit": None,
                    "type": MountPointType.HOSTPATH,
                },
                {
                    "local_path": "shared-data",
                    "mount_path": "/mnt/results/",
                    "read_only": False,
                    "size_limit": "10Gi",
                    "type": MountPointType.EMPTYDIR,
                },
            ],
            "resources": {
                "limits": {
                    "cpu": "1",
                    "memory": "5Gi",
                },
                "requests": {"cpu": "1", "memory": "512Mi"},
            },
            "ports": [],
        },
        "service_account_name": "simulation-service-account",
        "regcred_secret_name": "regcred",
        "node_affinity": {"key": "is_compute", "operator": "in", "values": ["true"]},
        "annotations": [],
        "labels": [],
    }



from unittest.mock import MagicMock, patch

import pytest
from simple_settings import LazySettings

from container_hub.carriers.kubernetes.backend import KubernetesBackend
from container_hub.models import KubernetesBackendConfig, KubernetesJobConfig


@pytest.fixture
def kubernetes_backend():
    config = KubernetesBackendConfig(
        "http://kubernetes/",
        "threedi",
    )
    return KubernetesBackend(config, in_cluster=False)


@pytest.fixture
def kubernetes_simple_settings():
    yield LazySettings("tests.test_files.kubernetes_settings")


def test_container_list(kubernetes_backend: KubernetesBackend):
    with patch("container_hub.carriers.kubernetes.backend.ApiClient") as client:
        job = MagicMock()
        job.metadata.name = "simulation-112"
        job_list = MagicMock()
        job_list.items = [job]
        job_list.metadata._continue = None
        client().__enter__().call_api.return_value = job_list
        hosts = kubernetes_backend.container_list()
        assert hosts == ["112"]


def get_job_status(kubernetes_backend: KubernetesBackend):
    with patch("container_hub.carriers.kubernetes.backend.ApiClient") as client:
        job = MagicMock()
        job.metadata.name = "simulation-112"
        job.status = MagicMock()
        job.status.active = 1
        job.status.succeeded = None
        job.status.failed = 1
        client().__enter__().call_api.return_value = job
        job_status = kubernetes_backend.get_job_status()
        assert job_status == {"active": 1, "succeeded": None, "failed": 1}


def test_pod_ip(kubernetes_backend: KubernetesBackend):
    with patch("container_hub.carriers.kubernetes.backend.ApiClient") as client:
        pod = MagicMock()
        pod.status.pod_ip = "127.0.0.1"
        pod_list = MagicMock()
        pod_list.items = [pod]
        client().__enter__().call_api.return_value = pod_list
        pod_ip = kubernetes_backend.pod_ip("simulation-1")
        assert pod_ip == "127.0.0.1"


def test_up(kubernetes_backend: KubernetesBackend, kubernetes_simple_settings):
    job_config = KubernetesJobConfig.from_settings(
        "simulation-1", kubernetes_simple_settings
    )

    with patch("container_hub.carriers.kubernetes.backend.ApiClient") as client:

        def call_api(*args, body=None, **kwargs):
            return (body, 200, {})

        client().__enter__().call_api = call_api
        name = kubernetes_backend.up(job_config)
        assert name == "simulation-1"


def test_down(kubernetes_backend: KubernetesBackend):
    with patch("container_hub.carriers.kubernetes.backend.ApiClient") as client:
        kubernetes_backend.down("simulation-1")
        client().__enter__().call_api.assert_called()



import logging
from functools import cached_property
from pathlib import Path
from typing import Dict, List

from docker import DockerClient
from docker.errors import APIError, ContainerError, DockerException, NotFound
from docker.types import Mount

from container_hub.exceptions import CarrierError
from container_hub.models import ContainerConfig, DockerBackendConfig

logger = logging.getLogger(__name__)


class DockerBackend:
    """
    Backend for starting Docker instances via Docker
    """

    def __init__(self, config: DockerBackendConfig):
        self.config = config

    @cached_property
    def client(self):
        return DockerClient(base_url=self.config.client_url)

    def container_hosts(self) -> Dict[str, str]:
        """
        Always localhost, so return emtpy dict
        """
        return {}

    def container_list(self) -> List[str]:
        """
        Returns a list of simulation_ids
        """
        lc = self.client.containers.list(filters={"label": "simulation_id"})
        return [x.name.lstrip("simulation-") for x in lc]

    def container_ips(self) -> Dict[str, str]:
        """
        Return list of ip addresses
        """
        d = {}
        containers = self.client.containers.list(filters={"label": "simulation_id"})
        for container in containers:
            try:
                ip_address = container.attrs["NetworkSettings"]["Networks"][
                    self.config.network_name
                ]["IPAddress"]
            except KeyError:
                continue
            d[container.name] = ip_address
        return d

    def up(self, container_config: ContainerConfig) -> str:
        """
        Create container based on simulation and threedimodel.

        :returns the container id
        """
        name = f"simulation-{container_config.sim_uid}"
        result_path = container_config.base_result_path / Path(name)
        labels = dict([(x.name, x.value) for x in container_config.labels])
        labels.update({"simulation_id": f"{container_config.sim_uid}"})

        _envs = dict([(f"{x.name}", f"{x.value}") for x in container_config.envs])
        _envs.update({"RESULT_PATH": f"{result_path.as_posix()}"})

        if (
            container_config.container_log_level is not None
            and "LOG_LEVEL" not in _envs
        ):
            _envs.update(
                {f"LOG_LEVEL": f"{container_config.container_log_level.value}"}
            )

        if container_config.debugpy_port is not None:
            ports = {
                f"{container_config.debugpy_port}/tcp": container_config.debugpy_port
            }
            _envs.update(
                {f"DEBUGPY": "1", "DEBUGPY_PORT": str(container_config.debugpy_port)}
            )
        else:
            ports = {}

        cmd = (
            f"python service.py {container_config.redis_host} {container_config.model_config} "
            f"{container_config.sim_uid} {container_config.sim_ref_datetime.isoformat()} "
            f"{container_config.end_time} {container_config.duration} {container_config.start_mode} "
            f"{container_config.pause_timeout} {container_config.max_rate} {container_config.clean_up_files}"
        )

        skip_model_mount: bool = all(
            [
                x
                for x in [
                    container_config.gridadmin_download_url,
                    container_config.tables_download_url,
                ]
            ]
        )

        if container_config.gridadmin_download_url is not None:
            cmd += f" {container_config.gridadmin_download_url}"
        if container_config.tables_download_url is not None:
            cmd += f" {container_config.tables_download_url}"

        logger.debug("cmd %s", cmd)
        logger.debug("Envs %s", _envs)

        mounts: List[Mount] = []
        for mount in container_config.mount_points:
            if skip_model_mount and mount.mount_path == "/models":
                # Skip mounting models
                continue
            mounts.append(
                Mount(
                    mount.mount_path,
                    mount.local_path,
                    type="bind",
                    read_only=mount.read_only,
                )
            )

        try:
            container = self.client.containers.run(
                image=container_config.image_name,
                command=cmd,
                name=name,
                network=self.config.network_name,
                mounts=mounts,
                environment=_envs,
                detach=True,
                labels=labels,
                ports=ports,
            )
        except (DockerException, APIError, ContainerError) as err:
            logger.error(err)
            raise CarrierError(err)

        # double check everything went right
        try:
            self.client.containers.get(container.id)
        except (APIError, NotFound) as err:
            logger.error(
                f"simulation container exited prematurely. Could not retrieve "
                f"the container if though it should be running {err}"
            )
            raise CarrierError(err)

        logger.info(f"Started simulation container {container.name}")
        return container.id

    def down(self, sim_uid: str):
        try:
            container = self.client.containers.get(f"simulation-{sim_uid}")
            container_id = container.id
        except (APIError, NotFound) as err:
            logger.error(
                f"Could not get the simulation container, error message: {err}"
            )
            raise CarrierError(err)
        try:
            if not self.config.debug:
                container.remove(force=True)
            else:
                container.kill()
        except APIError as err:
            logger.error(
                f"Could not kill/remove the "
                f"simulation container, error message: {err}"
            )
            raise CarrierError(err)
        logger.info(f"Removed container for simulation {sim_uid}")
        return container_id



CONTAINER_HUB_DEBUG = True
CONTAINER_HUB_CARRIER = "docker"
CONTAINER_HUB_CLIENT_URL = "unix://var/run/docker.sock"
CONTAINER_HUB_NETWORK_NAME = "threedi_backend"  # "threedi_backend"
CONTAINER_HUB_IMAGE_NAME = "testimage"
CONTAINER_HUB_CONTAINER_LOG_LEVEL = "DEBUG"
CONTAINER_HUB_MAX_CPU = 2

# harbor.lizard.net/threedi/threedicore:1.4.15.devtest_3"
CONTAINER_HUB_REDIS_HOST = "redis"  # "redis"
CONTAINER_HUB_BASE_MODEL_PATH = "/models"
# # Location where to write the results of the calccore.
CONTAINER_HUB_BASE_RESULT_PATH = "/results"


CONTAINER_HUB_MOUNT_POINTS = {
    "local_path_1": {"bind": "mount_path_1", "ro": True},
    "local_path_2": {"bind": "mount_path_2", "ro": False},
}



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

import codecs
import os
import re

from setuptools import find_packages, setup

here = os.path.abspath(os.path.dirname(__file__))


def read(*parts):
    with codecs.open(os.path.join(here, *parts), "r") as fp:
        return fp.read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


with open("README.rst") as readme_file:
    readme = readme_file.read()

with open("HISTORY.rst") as history_file:
    history = history_file.read()

requirements = [
    "marathon>=0.11.0",
    "docker<8.0",
    "python-slugify",
    "kubernetes>=21.7.0,<22.0",
    "hikaru>=0.11.0b0,<1.0",
]

setup_requirements = [
    "pytest-runner",
]

test_requirements = [
    "pytest>=3",
]

setup(
    author="Lars Claussen",
    author_email="claussen.lars@nelen-schuurmans.nl",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.8",
    ],
    description="The 3Di simulation container hub",
    install_requires=requirements,
    license="MIT license",
    long_description=readme + "\n\n" + history,
    long_description_content_type="text/x-rst",
    include_package_data=True,
    keywords="container_hub",
    name="container_hub",
    packages=find_packages(include=["container_hub", "container_hub.*"]),
    setup_requires=setup_requirements,
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/nens/container-hub",
    version=find_version("container_hub", "version.py"),
    zip_safe=False,
)



from datetime import datetime
from unittest.mock import MagicMock, patch

import pytest
from docker.types import Mount

from container_hub.carriers.docker.backend import DockerBackend
from container_hub.models import (
    ContainerConfig,
    DockerBackendConfig,
    EnvVar,
    Label,
    LogLevel,
    MountPoint,
)


@pytest.fixture
def docker_backend():
    config = DockerBackendConfig(
        "client_url",
        "my_network",
    )
    return DockerBackend(config)


def test_container_list(docker_backend: DockerBackend):
    with patch("container_hub.carriers.docker.backend.DockerClient") as client:
        container = MagicMock()
        container.name = "simulation-112"
        client().containers.list.return_value = [container]
        containers = docker_backend.container_list()
        assert containers == [
            "112",
        ]


def test_container_ips(docker_backend: DockerBackend):
    with patch("container_hub.carriers.docker.backend.DockerClient") as client:
        container = MagicMock()
        container.name = "simulation-112"
        container.attrs = {
            "NetworkSettings": {
                "Networks": {
                    docker_backend.config.network_name: {"IPAddress": "127.0.0.1"}
                }
            }
        }
        client().containers.list.return_value = [container]
        ip_addresses = docker_backend.container_ips()
        assert ip_addresses == {"simulation-112": "127.0.0.1"}


def test_up(docker_backend: DockerBackend):
    dt = datetime.now()
    container_config = ContainerConfig(
        "my_image",
        "base_result_path",
        12,
        dt,
        3600,
        3600,
        0,
        "initialize",
        "/model.ini",
        2,
        512,
        [EnvVar("env", "1")],
        [Label("name", "value")],
        0,
        True,
        "gridadmin_url",
        "tables_download_url",
        [MountPoint("/local", "/mnt", False)],
        "redis1",
        LogLevel.debug,
        5678,
    )
    with patch("container_hub.carriers.docker.backend.DockerClient") as client:
        container = MagicMock()
        container.id = 10
        client().containers.run.return_value = container
        container_id = docker_backend.up(container_config)
        assert container_id == 10

        # Check all params for DockerClient().containers.run
        to_check = {
            "image": "my_image",
            "command": f"python service.py redis1 /model.ini 12 {dt.isoformat()} 3600 3600 initialize 0 0 True gridadmin_url tables_download_url",
            "name": "simulation-12",
            "network": "my_network",
            "mounts": [
                Mount(
                    **{
                        "target": "/mnt",
                        "source": "/local",
                        "type": "bind",
                        "read_only": False,
                    }
                )
            ],
            "environment": {
                "env": "1",
                "RESULT_PATH": "base_result_path/simulation-12",
                "LOG_LEVEL": "DEBUG",
                "DEBUGPY": "1",
                "DEBUGPY_PORT": "5678",
            },
            "ports": {"5678/tcp": 5678},
            "detach": True,
            "labels": {"name": "value", "simulation_id": "12"},
        }
        assert client().containers.run.call_args[1] == to_check


def test_down(docker_backend: DockerBackend):
    with patch("container_hub.carriers.docker.backend.DockerClient") as client:
        container = MagicMock()
        container.id = 112
        client().containers.get.return_value = container
        container_id = docker_backend.down("112")
        assert container_id == 112
